// Fluxions 3D Engine
// shaderprogram.hpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#ifndef FLUXIONS_SHADERPROGRAM_HPP
#define FLUXIONS_SHADERPROGRAM_HPP

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <fluxions/opengl.hpp>
#include <fluxions/utils.hpp>
#include <fluxions/ftl_math.hpp>

namespace Fluxions
{
	struct Uniform
	{
		std::vector<float> float_vector;
		std::vector<int> int_vector;
		GLenum type;
		GLsizei count;

		Uniform();
		Uniform(const Uniform &u);
		~Uniform();

		const Uniform &operator = (const Uniform &u);
		float *GetFloatArray() { return &float_vector[0]; }
		float GetFloat() { return float_vector[0]; }
		int *GetIntegerArray() { return &int_vector[0]; }
		int GetInteger() { return int_vector[0]; }
		void operator=(const int &v);
		void operator=(const ftl::tvector2<int> &v);
		void operator=(const ftl::tvector3<int> &v);
		void operator=(const ftl::tvector4<int> &v);
		void operator=(const float &v);
		void operator=(const ftl::tvector2<float> &v);
		void operator=(const ftl::tvector3<float> &v);
		void operator=(const ftl::tvector4<float> &v);
		void operator=(const ftl::tmatrix2<float> &m);
		void operator=(const ftl::tmatrix3<float> &m);
		void operator=(const ftl::tmatrix4<float> &m);
		void operator=(const ftl::tmatrix2x3<float> &m);
		void operator=(const ftl::tmatrix2x4<float> &m);
		void operator=(const ftl::tmatrix3x2<float> &m);
		void operator=(const ftl::tmatrix3x4<float> &m);
		void operator=(const ftl::tmatrix4x2<float> &m);
		void operator=(const ftl::tmatrix4x3<float> &m);
		void IntegerVector(int count, int size, const int *v);
		void BoolVector(int count, int size, const int *v);
		void FloatVector(int count, int size, const float *v);
		void Matrix2(int count, const float *v);
		void Matrix3(int count, const float *v);
		void Matrix4(int count, const float *v);
		void Matrix2x3(int count, const float *v);
		void Matrix2x4(int count, const float *v);
		void Matrix3x2(int count, const float *v);
		void Matrix3x4(int count, const float *v);
		void Matrix4x2(int count, const float *v);
		void Matrix4x3(int count, const float *v);
		void Sampler1D(int count, int *v);
		void Sampler2D(int count, int *v);
		void Sampler3D(int count, int *v);
		void SamplerCube(int count, int *v);
		void Sampler1DShadow(int count, int *v);
		void Sampler2DShadow(int count, int *v);
		void SetUniform(GLint loc);
	};

	typedef std::map<std::string, Uniform> UniformLibrary;
	typedef std::map<std::string, Uniform>::iterator UniformLibraryIterator;

	class ShaderProgram
	{
	public:
		ShaderProgram();
		~ShaderProgram();

		bool Create();
		void Delete();
		void Use();
		bool Link();
		void BuildIndex();
		void PrintDetails(std::ostream &out);
		std::string GetLastErrorString();
		std::string GetShaderInfoLogString(GLuint shaderObj);
		std::string GetProgramInfoLogString(GLuint programObj);
		bool AttachShaderCode(GLenum type, const std::string &name, const std::string &code);
		bool AttachShader(GLenum type, const std::string &name, const std::string &filename);
		void DetachShader(const std::string &name);
		bool QuickCreate(const std::string &vertexShaderFilename, const std::string &fragmentShaderFilename);

		bool Load(const std::string &shaderName);
		void SetUniforms(UniformLibrary &uniforms);
		std::string TranslateAttribUniformName(const std::string &name);
		GLuint GetProgramId() { return programId; }
		int GetUniformIndex(const char *uniformName);
		int GetUniformIndex(const std::string &uniformName);
		int GetAttribIndex(const char *attribName);
		int GetAttribIndex(const std::string &attribName);
		void UniformVector1i(const std::string &uniformName, const int v);
		void UniformVector1iv(const std::string &uniformName, int count, const int *v);
		void UniformVector1f(const std::string &uniformName, const float v);
		void UniformVector1fv(const std::string &uniformName, int count, const float *v);
		void UniformVector2f(const std::string &uniformName, const ftl::Vector2f &vector);
		void UniformVector2fv(const std::string &uniformName, int count, const ftl::Vector2f *vector);
		void UniformVector2fv(const std::string &uniformName, int count, const float *v);
		void UniformVector3f(const std::string &uniformName, const ftl::Vector3f &vector);
		void UniformVector3fv(const std::string &uniformName, int count, const ftl::Vector3f *v);
		void UniformVector3fv(const std::string &uniformName, int count, const float *v);
		void UniformVector4f(const std::string &uniformName, const ftl::Vector4f &vector);
		void UniformVector4fv(const std::string &uniformName, int count, const ftl::Vector4f *vector);
		void UniformVector4fv(const std::string &uniformName, int count, const float *v);
		void UniformMatrix4f(const std::string &uniformName, const ftl::Matrix4f &matrix);
		void UniformMatrix4fv(const std::string &uniformName, int count, const ftl::Matrix4f *matrix);

		std::map<std::string, std::string> attribUniformNameMap;
	private:
		GLuint programId;
		std::string errorString;
		std::map<std::string, GLuint> shaderMap;
		std::map<std::string, int> uniformMap;
		std::map<std::string, int> attribMap;
	};
}

#endif
