#include <iostream>
#include "fluxions/fluxions_deferred_renderer.hpp"
#include "fluxions/shaderprogram.hpp"
#include "fluxions/Simple3DEngine.hpp"
#include "fluxions/glut_extensions.hpp"


namespace Fluxions
{

	SimpleDeferredRenderer::SimpleDeferredRenderer()
	{
		framebuffers[0] = 0;
		framebuffers[1] = 0;
		depthStencilTexture = 0;
		colorAttachmentTextures.clear();
	}


	SimpleDeferredRenderer::~SimpleDeferredRenderer()
	{
		if (framebuffers[0] != 0)
		{
			Delete();
		}
	}


	void SimpleDeferredRenderer::Create(int imageWidth_, int imageHeight_, GLenum depthStencilFormat, int numColorAttachments, GLenum colorAttachmentFormat)
	{
		if (framebuffers[0] != 0)
		{
			Delete();
		}

		// We use power of 2 textures to render...
		imageWidth = imageWidth_;
		imageHeight = imageHeight_;
		textureWidth = NextPowerOf2(imageWidth);
		textureHeight = NextPowerOf2(imageHeight);

		// Create textures...
		glActiveTexture(GL_TEXTURE0);
		depthStencilTexture = CreateTexture(GL_TEXTURE_2D, GL_DEPTH24_STENCIL8, textureWidth, textureHeight, GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8);
		colorAttachmentTextures.resize(numColorAttachments);
		for (int i = 0; i < numColorAttachments; i++)
		{
			colorAttachmentTextures[i] = CreateTexture(GL_TEXTURE_2D, GL_RGBA32F, textureWidth, textureHeight, GL_RGBA, GL_FLOAT);
		}

		glGenFramebuffers(3, framebuffers);
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[0]);
		glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, depthStencilTexture, 0);
		for (int i = 0; i < numColorAttachments; i++)
		{
			glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+i, GL_TEXTURE_2D, colorAttachmentTextures[i], 0);
		}
		if (CheckForCompleteness() != true)
		{
			cout << "SimpleDeferredRenderer::Create() -- failed to create framebuffer object 2" << endl;
		}
		glBindFramebuffer(GL_FRAMEBUFFER, 0);

		// Create Accumulation Framebuffer Object
		glGenTextures(2, accumBufferTextures);
		currentAccumBufferTexture = 0;
		glBindTexture(GL_TEXTURE_RECTANGLE, accumBufferTextures[0]);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA32F, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);
		glBindTexture(GL_TEXTURE_RECTANGLE, accumBufferTextures[1]);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA32F, textureWidth, textureHeight, 0, GL_RGBA, GL_FLOAT, 0);
		glBindTexture(GL_TEXTURE_RECTANGLE, 0);

		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[1]);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE, accumBufferTextures[0], 0);
		if (CheckForCompleteness() != true)
		{
			cout << "SimpleDeferredRenderer::Create() -- failed to create framebuffer object 2" << endl;
		}
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[2]);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_RECTANGLE, accumBufferTextures[1], 0);
		if (CheckForCompleteness() != true)
		{
			cout << "SimpleDeferredRenderer::Create() -- failed to create framebuffer object 3" << endl;
		}
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}


	void SimpleDeferredRenderer::Blit(int x, int y, int fbo)
	{
		glBindFramebuffer(GL_READ_FRAMEBUFFER, framebuffers[fbo]);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		glBlitFramebuffer(0, 0, textureWidth, textureHeight, x, y, x + textureWidth, y + textureHeight, GL_COLOR_BUFFER_BIT, GL_NEAREST);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}


	void SimpleDeferredRenderer::BlitGBuffer(int x, int y, int which)
	{
		glBindFramebuffer(GL_READ_FRAMEBUFFER, framebuffers[0]);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		if (which < 0) which = 0;
		if (which >= 0 && which < 4)
		{
			glReadBuffer(GL_COLOR_ATTACHMENT0+which);
			glBlitFramebuffer(0, 0, textureWidth, textureHeight, x, y, x + textureWidth, y + textureHeight, GL_COLOR_BUFFER_BIT, GL_NEAREST);
			glReadBuffer(GL_COLOR_ATTACHMENT0);
		}
		else if (which == 4)
		{
			glBlitFramebuffer(0, 0, textureWidth, textureHeight, x, y, x + textureWidth, y + textureHeight, GL_DEPTH_BUFFER_BIT, GL_NEAREST);
		}
		else if (which == 5)
		{
			glBlitFramebuffer(0, 0, textureWidth, textureHeight, x, y, x + textureWidth, y + textureHeight, GL_STENCIL_BUFFER_BIT, GL_NEAREST);
		}
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}


	void SimpleDeferredRenderer::BlitAccumBuffer(int x, int y, int which)
	{
		if (which < 0)
			glBindFramebuffer(GL_READ_FRAMEBUFFER, framebuffers[1+1 - currentAccumBufferTexture]);
		else
			glBindFramebuffer(GL_READ_FRAMEBUFFER, framebuffers[1+which]);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
		glBlitFramebuffer(0, 0, textureWidth, textureHeight, x, y, x + textureWidth, y + textureHeight, GL_COLOR_BUFFER_BIT, GL_NEAREST);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}


	bool SimpleDeferredRenderer::CheckForCompleteness()
	{
		GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
		if (status != GL_FRAMEBUFFER_COMPLETE)
		{
			cout << "SimpleDeferredRenderer::CheckForCompleteness() -- ";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE";
			if (status == GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER)
				cout << "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER";
			GLenum err = glGetError();
			if (err)
			{
				printf(" -- + GL error detected\n");
			}
			cout << endl;
			return false;
		}
		return true;
	}


	void SimpleDeferredRenderer::Delete()
	{
		if (framebuffers[0] != 0)
		{
			glDeleteFramebuffers(2, framebuffers);
			framebuffers[0] = 0;
			framebuffers[1] = 0;
		}
		
		if (depthStencilTexture != 0)
		{
			glDeleteTextures(1, &depthStencilTexture);
			depthStencilTexture = 0;
		}

		if (colorAttachmentTextures.empty() != false)
		{
			glDeleteTextures(colorAttachmentTextures.size(), &colorAttachmentTextures[0]);
			colorAttachmentTextures.clear();
		}
	}


	void SimpleDeferredRenderer::SetRenderingStage(DeferredRenderingStageTypeEnum stageType)
	{
		switch (stageType)
		{
		case DRST_NONE:
			glBindFramebuffer(GL_FRAMEBUFFER, 0);
			BindTextures(GL_TEXTURE0, true);
			break;
		case DRST_GBUFFER_PASS:
			// we want to render to our depth/stencil buffers and our color attachments...
			glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[0]);
			break;
		case DRST_LIGHT_PASS:
			// we want to render to the accumulation buffer using our color attachments as textures...
			glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[1]);
			BindTextures(GL_TEXTURE0, false);
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ONE);
			break;
		case DRST_COMBINE_PASS:
			break;
		case DRST_FINAL_PASS:
			break;
		default:
			// nothing to do...
			break;
		}
	}


	void SimpleDeferredRenderer::BindTextures(GLenum startTexture, bool unbind)
	{
		int colorAttachmentCount = colorAttachmentTextures.size();
		glActiveTexture(startTexture);
		glBindTexture(GL_TEXTURE_2D, depthStencilTexture);
		for (int i = 0; i < colorAttachmentCount; i++)
		{
			glActiveTexture(startTexture + i + 1);
			if (unbind != true)
				glBindTexture(GL_TEXTURE_2D, 0);
			else
				glBindTexture(GL_TEXTURE_2D, colorAttachmentTextures[i]);
		}
		glActiveTexture(GL_TEXTURE0);
	}


	int SimpleDeferredRenderer::NextPowerOf2(int x)
	{
		int y = 1;
		while (y < x)
		{
			y <<= 1;
		}
		return y;
	}


	GLuint SimpleDeferredRenderer::CreateTexture(GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels, GLint baseLevel, GLint maxLevel)
	{
		GLuint texture;
		glGenTextures(1, &texture);
		glBindTexture(target, texture);
		glTexImage2D(target, 0, internalformat, width, height, 0, format, type, pixels);
		glGenerateMipmap(target);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, baseLevel);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, maxLevel);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST_MIPMAP_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glBindTexture(target, 0);
		return texture;
	}


	void SimpleDeferredRenderer::SetupGBufferPass(const ftl::Matrix4f &pMatrix, const ftl::Matrix4f &vMatrix)
	{
		this->projectionMatrix = pMatrix;
		this->viewMatrix = vMatrix;

		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[0]);
		GLenum drawBufs[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3 };
		glDrawBuffers(4, drawBufs);
		glViewport(0, 0, textureWidth, textureHeight);
		glClearColor(0.0, 0.0, 0.0, 1.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);
	}


	void SimpleDeferredRenderer::RenderGBufferPass(Simple3DEngine &scenegraph)
	{
		UniformLibrary deferredUniforms;
		deferredUniforms["ViewMatrix"] = viewMatrix;
		deferredUniforms["ProjectionMatrix"] = projectionMatrix;
		deferredUniforms["ViewProjectionMatrix"] = projectionMatrix * viewMatrix;
		deferredUniforms["CameraPosition"] = ftl::Vector4f(0.0, 0.0, 0.0, 1.0) * viewMatrix;
		scenegraph.SetGlobalShaderOverride("deferred_gbuffer");
		scenegraph.SetGlobalShaderUniforms(deferredUniforms);
		scenegraph.Render();
		scenegraph.ClearState();
	}


	void SimpleDeferredRenderer::EndGBufferPass()
	{
		glDisable(GL_DEPTH_TEST);
	}


	void SimpleDeferredRenderer::ClearAccumBuffer()
	{
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[1+1 - currentAccumBufferTexture]);
		glClear(GL_COLOR_BUFFER_BIT);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}


	void SimpleDeferredRenderer::SetupDeferredLightingPass(ftl::Vector4f &lightPosition, ftl::Matrix4f &shadowMatrix, Simple3DEngine &scenegraph)
	{
		//this->lightPosition = lightPosition;
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffers[1+currentAccumBufferTexture]);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_RECTANGLE, accumBufferTextures[1 - currentAccumBufferTexture]);
		for (int i = 0; i < 4; i++)
		{
			glActiveTexture(GL_TEXTURE1+i);
			glBindTexture(GL_TEXTURE_2D, colorAttachmentTextures[i]);
		}
		glActiveTexture(GL_TEXTURE5);
		glBindTexture(GL_TEXTURE_2D, depthStencilTexture);
		currentAccumBufferTexture = 1 - currentAccumBufferTexture;

		ftl::Matrix4f orthoMatrix;
		orthoMatrix.LoadOrtho2D(-1.0, 1.0, -1.0, 1.0);
		//orthoMatrix.LoadOrtho2D(0, textureWidth, textureHeight, 0);

		UniformLibrary deferredUniforms;
		deferredUniforms["ViewMatrix"] = viewMatrix;
		deferredUniforms["ProjectionMatrix"] = projectionMatrix;
		deferredUniforms["ViewProjectionMatrix"] = orthoMatrix;
		deferredUniforms["CameraPosition"] = ftl::Vector4f(0.0, 0.0, 0.0, 1.0) * viewMatrix;
		deferredUniforms["LightPosition"] = lightPosition;
		deferredUniforms["ShadowMatrix"] = shadowMatrix;
		deferredUniforms["abufferTexture"] = 0;
		deferredUniforms["gbufferDiffuseTexture"] = 3;
		deferredUniforms["gbufferPositionTexture"] = 2;
		deferredUniforms["gbufferNormalTexture"] = 1;
		deferredUniforms["gbufferSpecularTexture"] = 4;
		deferredUniforms["gbufferDepthTexture"] = 5;
		deferredUniforms["LightShadowMap"] = 6;

		scenegraph.Shaders["deferred_lighting_pass"].Use();
		scenegraph.Shaders["deferred_lighting_pass"].SetUniforms(deferredUniforms);
		vloc = scenegraph.Shaders["deferred_lighting_pass"].GetAttribIndex("Vertex");
		tloc = scenegraph.Shaders["deferred_lighting_pass"].GetAttribIndex("TexCoord1");

		glViewport(0, 0, textureWidth, textureHeight);
		glClearColor(0.1, 0.2, 0.2, 1.0);
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
	}


	void SimpleDeferredRenderer::RenderDeferredLightingPass()
	{
		glutSpriteVertexArray(-1.0, -1.0, 1.0, 1.0, vloc, tloc, false, true);
		//glUseProgram(0);
		//glutRenderQuickOrthoRect();
		//glutRenderQuickTeapot();
	}


	void SimpleDeferredRenderer::EndDeferredLightingPass()
	{
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_RECTANGLE, 0);
		glClearColor(0.0, 0.0, 0.0, 1.0);
		glUseProgram(0);
	}

}
