#include <fluxions/lens_camera.hpp>
#include <fluxions/glut_extensions.hpp>


namespace Fluxions
{
	LensCamera::LensCamera()
	{
		clearColor = ftl::Vector4f(0.0, 0.0, 0.0, 1.0);
	}


	LensCamera::~LensCamera()
	{
	}


	void LensCamera::Init(int width, int height, float znear, float zfar, bool useDepthBuffer, GLenum depthBufferFunc)
	{
		screenWidth = width;
		screenHeight = height;
		this->znear = znear;
		this->zfar = zfar;
		this->useDepthBuffer = useDepthBuffer;
		this->depthBufferFunc = depthBufferFunc;
		QuickSet35mm(50.0);
		SetPositionOrientation(0.0f, 0.0f, 0.0f, 0.0f, 0.0f);
		RecalculateCamera();
	}


	void LensCamera::QuickSet4x5(float focalLength)
	{
		this->focalLength = focalLength;
		imageWidth = 101.6f; // 4" = 101.6mm
		imageHeight = 127.0f;  // 5" = 127.0mm
		coc = 0.06f;
		RecalculateCamera();
	}


	void LensCamera::QuickSetAPS(float focalLength)
	{
		this->focalLength = focalLength;
		imageWidth = 24.0f;
		imageHeight = 16.0f;
		coc = 0.02f;
		RecalculateCamera();
	}


	void LensCamera::QuickSet35mm(float focalLength)
	{
		this->focalLength = focalLength;
		imageWidth = 36.0f;
		imageHeight = 24.0f;
		coc = 0.03f;
		RecalculateCamera();
	}


	void LensCamera::QuickSetSuper65(float focalLength)
	{
		this->focalLength = focalLength;
		imageWidth = 65.0f;
		imageHeight = 24.0f;
		coc = 0.04f;
		RecalculateCamera();
	}


	void LensCamera::SetViewport(int screenWidth_, int screenHeight_)
	{
		screenWidth = screenWidth_;
		screenHeight = screenHeight_;
		RecalculateCamera();
	}


	void LensCamera::SetViewport()
	{
		glViewport(screenImageX, screenImageY, screenImageWidth, screenImageHeight);
	}


	void LensCamera::BeginFrame(GLbitfield clearBufferMask, GLenum direction)
	{
		savedViewport.Save();
		glViewport(screenImageX, screenImageY, screenImageWidth, screenImageHeight);
		if (clearBufferMask != 0)
		{
			glDisable(GL_DEPTH_TEST);
			glDepthMask(GL_FALSE);
			glColor4fv(clearColor.v);
			glutSprite(-1.0, -1.0, 1.0, 1.0);
			glDepthMask(GL_TRUE);
		}
		if (useDepthBuffer)
		{
			glEnable(GL_DEPTH_TEST);
			// TODO: Save old depth buffer function
			glDepthFunc(depthBufferFunc);
		}
	}


	void LensCamera::EndFrame()
	{
		savedViewport.Restore();
		if (useDepthBuffer)
		{
			glDisable(GL_DEPTH_TEST);
			// TODO: Restore old depth buffer function 
		}
	}


	void LensCamera::RecalculateCamera()
	{
		imageAspectRatio = imageWidth / imageHeight;
		screenImageWidth = (int)((float)screenHeight * imageAspectRatio);
		screenImageHeight = screenHeight;

		// Make sure it fits!
		if (screenImageWidth > screenWidth)
		{
			float scale = (float)screenWidth / (float)screenImageWidth;
			screenImageWidth = (int)((float)screenImageWidth * scale);
			screenImageHeight = (int)((float)screenImageHeight * scale);
		}

		// Center it
		screenImageX = (screenWidth - screenImageWidth) / 2;
		screenImageY = (screenHeight - screenImageHeight) / 2;

		// Now set up the lens
		lensMatrix.LoadIdentity();
		lensMatrix.LensMatrix(focalLength, imageWidth, imageAspectRatio);

		projectionMatrix.LoadIdentity();
		projectionMatrix.Perspective(znear, zfar);

		viewProjectionMatrix.LoadIdentity();
		viewProjectionMatrix = projectionMatrix * lensMatrix * worldMatrix.Inverse();

		worldPosition = ftl::Vector4f(0.0f, 0.0f, 0.0f, 1.0f) * worldMatrix;
	}
	

	void LensCamera::SetWorldPosition(float x, float y, float z)
	{
		worldMatrix.Translate(-worldPosition.x, -worldPosition.y, -worldPosition.z);
		worldMatrix.Translate(x, y, z);
		worldPosition.x = x;
		worldPosition.y = y;
		worldPosition.z = z;
		worldPosition.w = 1;
		RecalculateCamera();
	}


	void LensCamera::SetPositionOrientation(GLfloat x, GLfloat y, GLfloat z, GLfloat theta, GLfloat phi)
	{
		worldMatrix.LoadIdentity();
		worldMatrix.Translate(x, y, z);
		worldMatrix.Rotate(theta, 0.0f, 1.0f, 0.0f);
		worldMatrix.Rotate(phi, 1.0f, 0.0f, 0.0f);
		RecalculateCamera();
	}


	void LensCamera::SetLookAt(float x, float y, float z, float upx, float upy, float upz)
	{
		ftl::Vector3f forward(x, y, z);
		ftl::Vector3f up(upx, upy, upz);
		ftl::Vector3f right = ftl::CrossProduct(forward, up);
		up = ftl::CrossProduct(forward, right);
		forward.Normalize();
		up.Normalize();
		right.Normalize();		
		worldMatrix = ftl::Matrix3f(right, up, forward);
		worldMatrix.Translate(worldPosition.x, worldPosition.y, worldPosition.z);
		RecalculateCamera();
	}


	void LensCamera::SetLookAt(float theta, float phi)
	{
		worldMatrix.LoadIdentity();
		worldMatrix.Translate(worldPosition.x, worldPosition.y, worldPosition.z);
		worldMatrix.Rotate(theta, 0.0f, 1.0f, 0.0f);
		worldMatrix.Rotate(phi, 1.0f, 0.0f, 0.0f);
		RecalculateCamera();
	}


	void LensCamera::MoveRotate(float rx, float ry, float rz, float tx, float ty, float tz)
	{
		worldMatrix.Rotate(rx, 1.0f, 0.0f, 0.0f);
		worldMatrix.Rotate(ry, 0.0f, 1.0f, 0.0f);
		worldMatrix.Rotate(rz, 0.0f, 0.0f, 1.0f);
		worldMatrix.Translate(tx, ty, tz);
		RecalculateCamera();
	}


	void LensCamera::SetClearColor(const float *rgba)
	{
		clearColor = rgba;
	}


	void LensCamera::SetClearColor(float r, float g, float b, float a)
	{
		clearColor.r = r;
		clearColor.g = g;
		clearColor.b = b;
		clearColor.a = a;
	}
}
