#include <libmfg_kasl.hpp>
#include <ctype.h>

using namespace std;


KASL_VarCmdProcessor::KASL_VarCmdProcessor()
{
}


KASL_VarCmdProcessor::~KASL_VarCmdProcessor()
{
}


bool KASL_VarCmdProcessor::IsVariable(const char *str)
{
	return (GetVariable(str) == NULL);
}


KASL_Variable *KASL_VarCmdProcessor::GetVariable(const char *str)
{
	list<KASL_Variable>::iterator it;
	for (it = vars.begin(); it != vars.end(); it++)
	{
		if (it->name == str)
			return &(*it);
	}
	return NULL;
}

const char *KASL_VarCmdProcessor::PartialMatchVariableName(const char *str)
{
	list<KASL_Variable>::iterator it;
	for (it = vars.begin(); it != vars.end(); it++)
	{
		if (it->name.find(str) == 0)
			return it->name.c_str();
	}
	return NULL;
}
	
void KASL_VarCmdProcessor::AddVariable(const char *str, float value)
{
	KASL_Variable *var = GetVariable(str);
	if (var == NULL)
	{
		KASL_Variable v;
		v.name = str;
		v.value = value;
		vars.push_back(v);
	}
	else
	{
		var->value = value;
	}
}


void KASL_VarCmdProcessor::AddVariable(const char *str, const char *string)
{
	KASL_Variable *var = GetVariable(str);
	if (var == NULL)
	{
		KASL_Variable v;
		v.name = str;
		v.string = string;
		vars.push_back(v);
	}
	else
	{
		var->string = string;
	}
}
	
void KASL_VarCmdProcessor::SetVariableValue(const char *str, float value)
{
	KASL_Variable *var = GetVariable(str);
	if (var == NULL)
	{
		KASL_Variable v;
		v.name = str;
		v.value = value;
		vars.push_back(v);
	}
	else
	{
		var->value = value;
	}
}


void KASL_VarCmdProcessor::SetVariableValue(const char *str, const char *string)
{
	KASL_Variable *var = GetVariable(str);
	float val = atof(string);
	if (var == NULL)
	{
		KASL_Variable v;
		v.name = str;
		v.string = string;
		v.value = val;
		vars.push_back(v);
	}
	else
	{
		var->string = string;
		var->value = val;
	}
}


void KASL_VarCmdProcessor::DeleteVariable(const char *str)
{
	list<KASL_Variable>::iterator it;
	for (it = vars.begin(); it != vars.end(); it++)
	{
		if (it->name == str)
			break;
	}
	if (it != vars.end())
	{
		vars.erase(it);
	}
}


float KASL_VarCmdProcessor::GetVariableValue(const char *str)
{
	KASL_Variable *var = GetVariable(str);
	if (var != NULL)
	{
		return var->value;
	}
	else
	{
		return 0;
	}
}


std::string KASL_VarCmdProcessor::GetVariableString(const char *str)
{
	KASL_Variable *var = GetVariable(str);
	if (var != NULL)
	{
		return var->string;
	}
	else
	{
		return "";
	}
}


void KASL_VarCmdProcessor::TokenizeString(const char *str, vector<string> &tokens)
{
	string tmp = str;
	string token;
	int len = tmp.length();
	for (int i = 0; i < len; i++)
	{
		if (isspace(tmp[i]))
		{
			if (!token.empty())
			{
				tokens.push_back(token);
				token = "";
			}
			continue;
		}
		token.push_back(tmp[i]);
	}
	if (!token.empty())
	{
		tokens.push_back(token);
	}
}


bool KASL_VarCmdProcessor::IsCommand(const char *str)
{
	return (GetCommand(str) != NULL);
}

KASL_Command *KASL_VarCmdProcessor::GetCommand(const char *str)
{
	list<KASL_Command>::iterator it;
	for (it = cmds.begin(); it != cmds.end(); it++)
	{
		if (it->name == str)
			return &(*it);
	}
	return NULL;
}

const char *KASL_VarCmdProcessor::PartialMatchCommandName(const char *str)
{
	list<KASL_Command>::iterator it;
	for (it = cmds.begin(); it != cmds.end(); it++)
	{
		if (it->name.find(str) == 0)
			return it->name.c_str();
	}
	return NULL;
}

void KASL_VarCmdProcessor::AddCommand(const char *str, void (*func)(std::vector<std::string> &, void*), void *data)
{
	KASL_Command *cmd = GetCommand(str);
	if (cmd == NULL)
	{
		KASL_Command tmp;
		tmp.name = str;
		tmp.callback = func;
		tmp.data = data;
		cmds.push_back(tmp);
	}
	else
	{
		cmd->callback = func;
		cmd->data = data;
	}
}

void KASL_VarCmdProcessor::DeleteCommand(const char *str)
{
	list<KASL_Command>::iterator it;
	for (it = cmds.begin(); it != cmds.end(); it++)
	{
		if (it->name == str)
			break;
	}
	if (it != cmds.end())
	{
		cmds.erase(it);
	}
}

bool KASL_VarCmdProcessor::Execute(const char *str)
{
	vector<string> cmdList;
	TokenizeString(str, cmdList);
	if (cmdList.empty())
		return false;
	KASL_Command *cmd = GetCommand(cmdList[0].c_str());
	if (cmd != NULL)
	{
		if (cmd->callback == NULL)
			return false;
		cmd->callback(cmdList, cmd->data);
		return true;
	}
	else
	{
		// assign a variable
		if (cmdList.size() >= 2)
			SetVariableValue(cmdList[0].c_str(), cmdList[1].c_str());
		return false;
	}
}
