// Fluxions 3D Engine
// shaderprogram.cpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#include <fluxions/shaderprogram.hpp>
#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <sstream>

namespace Fluxions
{
	using namespace std;
	using namespace ftl;

	const char *GLEnumToString(GLenum v)
	{
		switch (v)
		{
#define CASESTRING(x)	case x: return #x;
			CASESTRING(GL_FLOAT);
			CASESTRING(GL_FLOAT_VEC2);
			CASESTRING(GL_FLOAT_VEC3);
			CASESTRING(GL_FLOAT_VEC4);
			CASESTRING(GL_INT);
			CASESTRING(GL_INT_VEC2);
			CASESTRING(GL_INT_VEC3);
			CASESTRING(GL_INT_VEC4);
			CASESTRING(GL_BOOL);
			CASESTRING(GL_BOOL_VEC3);
			CASESTRING(GL_BOOL_VEC4);
			CASESTRING(GL_FLOAT_MAT2);
			CASESTRING(GL_FLOAT_MAT3);
			CASESTRING(GL_FLOAT_MAT4);
			CASESTRING(GL_FLOAT_MAT2x3);
			CASESTRING(GL_FLOAT_MAT2x4);
			CASESTRING(GL_FLOAT_MAT3x4);
			CASESTRING(GL_FLOAT_MAT4x2);
			CASESTRING(GL_FLOAT_MAT4x3);
			CASESTRING(GL_SAMPLER_1D);
			CASESTRING(GL_SAMPLER_2D);
			CASESTRING(GL_SAMPLER_3D);
			CASESTRING(GL_SAMPLER_CUBE);
			CASESTRING(GL_SAMPLER_1D_SHADOW);
			CASESTRING(GL_SAMPLER_2D_SHADOW);
#undef CASESTRING
		}
		return "<unknown>";
	}

	ShaderProgram::ShaderProgram()
	: programId(0)
	{
		
	}
	
	ShaderProgram::~ShaderProgram()
	{
		
	}
	
	bool ShaderProgram::Create()
	{
		programId = glCreateProgram();
		if (programId == 0)
			return false;
		return true;
	}
	
	void ShaderProgram::Delete()
	{
		if (programId != 0)
		{
			glDeleteProgram(programId);
			map<string, GLuint>::iterator it;
			for (it = shaderMap.begin(); it != shaderMap.end(); ++it)
			{
				glDeleteShader(it->second);
			}
			shaderMap.clear();
		}
	}
	
	void ShaderProgram::Use()
	{
		glUseProgram(programId);
	}
	
	bool ShaderProgram::Link()
	{
		GLint status;
		uniformMap.clear();
		glLinkProgram(programId);
		glGetProgramiv(programId, GL_LINK_STATUS, &status);
		cout << "Program (" << programId << ") Info Log:\n" << GetProgramInfoLogString(programId) << "\n\n";
		if (status == GL_FALSE)
		{
			errorString = "Error Linking";
			return false;
		}
		BuildIndex();
		PrintDetails(cout);
		return true;
	}

	void ShaderProgram::BuildIndex()
	{
		// rebuild uniforms and attribs
		int numActiveUniforms;
		int numActiveAttribs;
		char *buffer;
		int length;
		GLsizei rlength;
		GLint rsize;
		GLenum rtype;

		glGetProgramiv(programId, GL_ACTIVE_UNIFORMS, &numActiveUniforms);
		glGetProgramiv(programId, GL_ACTIVE_UNIFORM_MAX_LENGTH, &length);
		if (length > 0)
		{
			buffer = (char *)malloc(length);
			if (!buffer) return;
			for (int i = 0; i < numActiveUniforms; ++i)
			{
				glGetActiveUniform(programId, i, length, &rlength, &rsize, &rtype, buffer);
				uniformMap[string(buffer)] = glGetUniformLocation(programId, buffer);
			}
			free(buffer);
		}

		glGetProgramiv(programId, GL_ACTIVE_ATTRIBUTES, &numActiveAttribs);
		glGetProgramiv(programId, GL_ACTIVE_ATTRIBUTE_MAX_LENGTH, &length);
		if (length > 0)
		{
			buffer = (char *)malloc(length);
			if (!buffer) return;
			for (int i = 0; i < numActiveAttribs; ++i)
			{
				glGetActiveAttrib(programId, i, length, &rlength, &rsize, &rtype, buffer);				
				attribMap[string(buffer)] = glGetAttribLocation(programId, buffer);
			}
			cout << endl;
			free(buffer);
		}
	}

	void ShaderProgram::PrintDetails(ostream &out)
	{
		// rebuild uniforms
		int numActiveUniforms;
		int numActiveAttribs;
		char *buffer;
		int length;
		GLsizei rlength;
		GLint rsize;
		GLenum rtype;

		glGetProgramiv(programId, GL_ACTIVE_UNIFORMS, &numActiveUniforms);
		glGetProgramiv(programId, GL_ACTIVE_UNIFORM_MAX_LENGTH, &length);
		if (length > 0)
		{
			buffer = (char *)malloc(length);
			if (!buffer) return;
			cout << "Program (" << programId << ") Active Uniforms: \n";
			for (int i = 0; i < numActiveUniforms; ++i)
			{
				glGetActiveUniform(programId, i, length, &rlength, &rsize, &rtype, buffer);
				int loc = glGetUniformLocation(programId, buffer);
				cout << "    " << i << "(" << loc << ")";
				cout << " " << setw(35) << GLEnumToString(rtype);
				cout << " " << string(buffer);
				if (rsize > 1)
					cout << "[" << rsize << "]\n";
				else
					cout << "\n";
			}
			free(buffer);
		}

		glGetProgramiv(programId, GL_ACTIVE_ATTRIBUTES, &numActiveAttribs);
		glGetProgramiv(programId, GL_ACTIVE_ATTRIBUTE_MAX_LENGTH, &length);
		if (length > 0)
		{
			buffer = (char *)malloc(length);
			if (!buffer) return;
			cout << "Program (" << programId << ") Active Attribs: \n";
			for (int i = 0; i < numActiveAttribs; ++i)
			{
				glGetActiveAttrib(programId, i, length, &rlength, &rsize, &rtype, buffer);
				int loc = glGetAttribLocation(programId, buffer);
				cout << "    " << i << "(" << loc << ")";
				cout << "    " << i;
				cout << " " << setw(20) << GLEnumToString(rtype);
				cout << " " << string(buffer);
				if (rsize > 1)
					cout << "[" << rsize << "]\n";
				else
					cout << "\n";
			}
			cout << endl;
			free(buffer);
		}
	}
	
	string ShaderProgram::GetLastErrorString()
	{
		return errorString;
	}
	
	string ShaderProgram::GetShaderInfoLogString(GLuint shaderObj)
	{
		string str;
		
		int infoLogLength;
		char *infoLog;
		glGetShaderiv(shaderObj, GL_INFO_LOG_LENGTH, &infoLogLength);
		if (infoLogLength == 0)
			return "";
		infoLog = new char[infoLogLength];
		glGetShaderInfoLog(shaderObj, infoLogLength, NULL, infoLog);
		str = infoLog;
		delete [] infoLog;
		return str;
	}
	
	string ShaderProgram::GetProgramInfoLogString(GLuint programObj)
	{
		string str;
		
		int infoLogLength;
		char *infoLog;
		glGetProgramiv(programObj, GL_INFO_LOG_LENGTH, &infoLogLength);
		if (infoLogLength == 0)
			return "";
		infoLog = new char[infoLogLength];
		glGetProgramInfoLog(programObj, infoLogLength, NULL, infoLog);
		str = infoLog;
		delete [] infoLog;
		return str;
	}

	bool ShaderProgram::AttachShaderCode(GLuint type, const string &name, const string &code)
	{
		GLuint shaderId;
		shaderId = glCreateShader(type);
		if (shaderId == 0)
			return false;
		const GLchar *sourcePtr = code.c_str();
		glShaderSource(shaderId, 1, &sourcePtr, NULL);
		glCompileShader(shaderId);
		GLint status;
		glGetShaderiv(shaderId, GL_COMPILE_STATUS, &status);
		cout << "Shader Log(" << name << "): " << GetShaderInfoLogString(shaderId) << endl;
		if (status == GL_FALSE)
		{
			glDeleteShader(shaderId);
			return false;
		}
		glAttachShader(programId, shaderId);
		shaderMap[name] = shaderId;
		return true;
	}
	
	bool ShaderProgram::AttachShader(GLuint type, const string &name, const string &filename)
	{
		string source;
		source = LoadEntireTextFile(filename);
		GLuint shaderId;
		shaderId = glCreateShader(type);
		if (shaderId == 0)
			return false;
		source.append(" ");
		const GLchar *sourcePtr = source.c_str();
		glShaderSource(shaderId, 1, &sourcePtr, NULL);
		glCompileShader(shaderId);
		GLint status;
		glGetShaderiv(shaderId, GL_COMPILE_STATUS, &status);
		cout << "Shader Log(" << filename << "): " << GetShaderInfoLogString(shaderId) << endl;
		if (status == GL_FALSE)
		{
			glDeleteShader(shaderId);
			return false;
		}
		glAttachShader(programId, shaderId);
		shaderMap[name] = shaderId;
		return true;
	}
	
	void ShaderProgram::DetachShader(const string &name)
	{
		map<string, GLuint>::iterator it;
		it = shaderMap.find(name);
		if (it == shaderMap.end())
			return;
		else
			shaderMap.erase(it);
	}
	
	bool ShaderProgram::QuickCreate(const string &vertexShaderFilename, const string &fragmentShaderFilename)
	{
		Delete();
		Create();
		cout << "Creating new program(" << programId << ")" << ": \n";
		if (!AttachShader(GL_VERTEX_SHADER, StripFileExtension(vertexShaderFilename), vertexShaderFilename))
		{
			Delete();
			return false;
		}
		if (!AttachShader(GL_FRAGMENT_SHADER, StripFileExtension(fragmentShaderFilename), fragmentShaderFilename))
		{
			Delete();
			return false;
		}
		return Link();
	}


	bool ShaderProgram::Load(const string &filename)
	{
		Delete();
		Create();
		cout << "=== ShaderProgram::Load() === START OF SHADER PROGRAM =================\n\n";
		cout << "Creating new program(" << programId << ":" << filename << ")" << ": \n\n";

		int vshadercount = 0;
		int fshadercount = 0;
		int gshadercount = 0;
		string currentshader;
		string shadername;
		stringstream sstr;
		string line;
		string str, tmp;
		ifstream fin(filename);
		const int NO_MODE = 0;
		const int ATTRIBUTE_MODE = 1;
		const int UNIFORM_MODE = 2;
		const int VERTEXSHADER_MODE = 3;
		const int GEOMETRYSHADER_MODE = 4;
		const int FRAGMENTSHADER_MODE = 5;
		int mode = NO_MODE;
		while (1)
		{
			getline(fin, line);
			if (!fin)
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				break;
			}
			sstr.clear();
			sstr.str(line);
			sstr >> str;
			if (str == "#")
			{
				// skip "# Comments"
				// but not "#preprocessor commands"
				continue;
			}
			else if (str == "@NAMES" || str == "@ATTRIBUTES")
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				mode = ATTRIBUTE_MODE;
				currentshader.clear();
			}
			else if (str == "@UNIFORMS")
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				mode = UNIFORM_MODE;
				currentshader.clear();
			}
			else if (str == "@VERTEXSHADER")
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				mode = VERTEXSHADER_MODE;
				shadername = "";
				sstr >> shadername;
				if (shadername == "")
					shadername = ToString256("VERTEXSHADER%d", vshadercount);
				currentshader.clear();
				vshadercount++;
			}
			else if (str == "@FRAGMENTSHADER")
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				mode = FRAGMENTSHADER_MODE;
				shadername = "";
				sstr >> shadername;
				if (shadername == "")
					shadername = ToString256("FRAGMENTSHADER%d", fshadercount);
				currentshader.clear();
				fshadercount++;
			}
			else if (str == "@GEOMETRYSHADER")
			{
				if (mode == VERTEXSHADER_MODE) { AttachShaderCode(GL_VERTEX_SHADER, shadername, currentshader); }
				if (mode == FRAGMENTSHADER_MODE) { AttachShaderCode(GL_FRAGMENT_SHADER, shadername, currentshader); }
				if (mode == GEOMETRYSHADER_MODE) { AttachShaderCode(GL_GEOMETRY_SHADER, shadername, currentshader); }
				mode = GEOMETRYSHADER_MODE;
				shadername = "";
				sstr >> shadername;
				if (shadername == "")
					shadername = ToString256("GEOMETRYSHADER%d", gshadercount);
				currentshader.clear();
				gshadercount++;
			}
			else
			{
				if (mode == VERTEXSHADER_MODE || mode == GEOMETRYSHADER_MODE || mode == FRAGMENTSHADER_MODE)
				{
					currentshader += line;
					currentshader += "\n";
				}
				else if (mode == ATTRIBUTE_MODE || UNIFORM_MODE)
				{
					sstr >> tmp;
					attribUniformNameMap[str] = tmp;
				}
			}
		}
		fin.close();

		cout << "Name Translations:\n";
		for (auto it = attribUniformNameMap.begin(); it != attribUniformNameMap.end(); it++)
		{
			cout << "    " << setw(35) << it->first << " ==> " << it->second << endl;
		}

		bool status = Link();
		cout << "--- ShaderProgram::Load() --- END OF SHADER PROGRAM -------------------\n\n";
		return status;
	}
	
	int ShaderProgram::GetUniformIndex(const char *uniformName)
	{
		int index = -1;
		map<string, int>::iterator it;
		it = uniformMap.find(uniformName);
		if (it != uniformMap.end())
			return it->second;
		return index;
	}
	
	int ShaderProgram::GetUniformIndex(const string &uniformName)
	{
		int index = -1;
		map<string, int>::iterator it;
		it = uniformMap.find(uniformName);
		if (it != uniformMap.end())
		{
			int loc = glGetUniformLocation(programId, uniformName.c_str());
			if (loc != it->second)
				cout << "aha!\n";
			return it->second;
		}
		return index;
	}

	int ShaderProgram::GetAttribIndex(const char *attribName)
	{
		int index = -1;
		map<string, int>::iterator it;
		it = attribMap.find(attribName);
		if (it != attribMap.end())
			return it->second;
		return index;
	}
	
	int ShaderProgram::GetAttribIndex(const string &attribName)
	{
		int index = -1;
		map<string, int>::iterator it;
		it = attribMap.find(attribName);
		if (it != attribMap.end())
			return it->second;
		return index;
	}

	void ShaderProgram::UniformVector1iv(const string &uniformName, int count, const int *v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform1iv(index, count, v);
	}

	void ShaderProgram::UniformVector1i(const string &uniformName, const int v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform1i(index, v);
	}

	void ShaderProgram::UniformVector1f(const string &uniformName, const float v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform1f(index, v);
	}

	void ShaderProgram::UniformVector1fv(const string &uniformName, int count, const float *v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform1fv(index, count, v);
	}

	void ShaderProgram::UniformVector2f(const string &uniformName, const Vector2f &vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform2fv(index, 1, vector.v);
	}

	void ShaderProgram::UniformVector2fv(const string &uniformName, int count, const Vector2f *vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform2fv(index, count, vector->v);
	}

	void ShaderProgram::UniformVector2fv(const string &uniformName, int count, const float *v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform2fv(index, count, v);
	}

	void ShaderProgram::UniformVector3f(const string &uniformName, const Vector3f &vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform3fv(index, 1, vector.v);
	}

	void ShaderProgram::UniformVector3fv(const string &uniformName, int count, const Vector3f *vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform3fv(index, count, vector->v);
	}

	void ShaderProgram::UniformVector3fv(const string &uniformName, int count, const float *v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform3fv(index, count, v);
	}

	void ShaderProgram::UniformVector4f(const string &uniformName, const Vector4f &vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform4fv(index, 1, vector.v);
	}

	void ShaderProgram::UniformVector4fv(const string &uniformName, int count, const Vector4f *vector)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform4fv(index, count, vector->v);
	}

	void ShaderProgram::UniformVector4fv(const string &uniformName, int count, const float *v)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniform4fv(index, count, v);
	}

	void ShaderProgram::UniformMatrix4f(const string &uniformName, const Matrix4f &matrix)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniformMatrix4fv(index, 1, GL_FALSE, matrix.m);
	}

	void ShaderProgram::UniformMatrix4fv(const string &uniformName, int count, const Matrix4f *matrix)
	{
		int index = GetUniformIndex(uniformName);
		if (index < 0)
			return;
		glUniformMatrix4fv(index, count, GL_FALSE, matrix->m);
	}


	void ShaderProgram::SetUniforms(UniformLibrary &uniforms)
	{
		if (programId == 0)
			return;
		UniformLibraryIterator it;
		for (it = uniforms.begin(); it != uniforms.end(); it++)
		{
			// do variable name translation lookup
			int loc = GetUniformIndex(TranslateAttribUniformName(it->first));
			if (loc < 0)
				continue;
			it->second.SetUniform(loc);
		}
	}

	// Class:  ShaderProgram
	// Method: string TranslateAttribUniformName(name)
	//
	// name - name of attribute or uniform that may have a different meaning
	//
	// Returns the actual name of the attribute or uniform used in the shader.
	// If there is no lookup in the database, then the original name is returned
	// since there is no need for translation or it doesn't exist.
	string ShaderProgram::TranslateAttribUniformName(const string &name)
	{
		map<string, string>::iterator it = attribUniformNameMap.find(name);
		if (it != attribUniformNameMap.end())
		{
			return it->second;
		}
		else
		{
			return name;
		}
	}
}
