// Fluxions 3D Engine
// staticmodel.cpp
// Copyright (C) 2000-2011 Jonathan Metzgar
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iomanip>
#include <string>
#include <map>
#include <fluxions/utils.hpp>
#include "fluxions/staticmodel.hpp"

using namespace std;

namespace Fluxions
{
	using namespace ftl;


	StaticModel::StaticModel()
		: arrayBufferID(0), elementArrayBufferID(0)
	{
		vertexBinding = 0;
		normalBinding = 0;
		texcoord1Binding = 0;
		texcoord2Binding = 0;
		texcoord3Binding = 0;
	}


	StaticModel::~StaticModel()
	{
	}


	bool StaticModel::Save(const std::string &filename)
	{
		int size;
		char fileID[] = "FLUXIONSMODEL";
		int fileVersion = 1;
		FILE *fout = fopen(filename.c_str(), "wb");
		if (!fout)
			return false;
		// Write Header
		fwrite(fileID, 1, 13, fout);
		fwrite(&fileVersion, 1, sizeof(int), fout);
		size = Vertices.size();
		fwrite(&size, 1, sizeof(int), fout);
		size = Indices.size();
		fwrite(&size, 1, sizeof(int), fout);
		size = Surfaces.size();
		fwrite(&size, 1, sizeof(int), fout);

		// Write Data
		fwrite(&Vertices[0], Vertices.size(), sizeof(Vertex64), fout);
		fwrite(&Indices[0], Indices.size(), sizeof(GLushort), fout);
		fwrite(&Surfaces[0], Surfaces.size(), sizeof(Surface), fout);

		fclose(fout);
		return true;
	}


	bool StaticModel::Load(const std::string &filename)
	{
		int size1, size2, size3;
		char fileID[] = "FLUXIONSMODEL";
		int fileVersion = 1;
		FILE *fin = fopen(filename.c_str(), "rb");
		fread(fileID, 1, 13, fin);
		if (strcmp(fileID, "FLUXIONSMODEL") != 0)
		{
			fclose(fin);
			return false;
		}
		fread(&fileVersion, 1, sizeof(int), fin);
		if (fileVersion != 1)
		{
			fclose(fin);
			return false;
		}

		fread(&size1, 1, sizeof(int), fin);
		fread(&size2, 1, sizeof(int), fin);
		fread(&size3, 1, sizeof(int), fin);
		Resize(size1, size2, size3);
		fread(&Vertices[0], Vertices.size(), sizeof(Vertex64), fin);
		fread(&Indices[0], Indices.size(), sizeof(GLushort), fin);
		fread(&Surfaces[0], Surfaces.size(), sizeof(Surface), fin);

		fclose(fin);
		return true;
	}


	bool StaticModel::LoadXYZ(const std::string &filename, const float size)
	{
		std::cout << "StaticModel::LoadXYZ() --- Loading " << filename << "\n";
		FILE *fin = fopen(filename.c_str(), "r");
		if (!fin)
			return false;

		int numVertices = 0, numTriangles = 0;
		float x, y, z;

		fscanf(fin, "%d %d", &numVertices, &numTriangles);
		if (numTriangles == 0)
			CreateSimpleModel(numVertices, numVertices, 1);
		else
			CreateSimpleModel(numVertices, numTriangles*3, 1);
		
		BoundingBox.Reset();
		for (int i = 0; i < numVertices; i++)
		{
			if ((i % (numVertices/10)) == 0)
				std::cout << "#";
			fscanf(fin, "%f %f %f", &x, &y, &z);
			Vertices[i].position.x = x;
			Vertices[i].position.y = y;
			Vertices[i].position.z = z;
			Vertices[i].normal = Vector3f(0, 0, 0);
			BoundingBox += Vector3f(Vertices[i].position);
		}
		std::cout << "\n";

		if (numTriangles != 0)
		{
			int index[3];
			for (int i = 0; i < numTriangles; i++)
			{
				fscanf(fin, "%d %d %d", &index[0], &index[1], &index[2]);
				Indices[i*3+0] = index[0];
				Indices[i*3+1] = index[1];
				Indices[i*3+2] = index[2];
				// Add Normal for polygon
				Vector3f p1 = Vertices[index[0]].position;
				Vector3f p2 = Vertices[index[1]].position;
				Vector3f p3 = Vertices[index[2]].position;
				Vector3f u = p2-p1;
				Vector3f v = p3-p2;
				Vector3f n = CrossProduct(u, v);
				n.Normalize();
				Vertices[index[0]].normal += n;
				Vertices[index[1]].normal += n;
				Vertices[index[2]].normal += n;
			}
		}
		else
		{
			numTriangles = Indices.size() / 3;
			int index[3];
			for (int i = 0; i < numTriangles; i++)
			{
				//fscanf(fin, "%d %d %d", &index[0], &index[1], &index[2]);
				index[0] = i*3+0;
				index[1] = i*3+1;
				index[2] = i*3+2;
				Indices[i*3+0] = index[0];
				Indices[i*3+1] = index[1];
				Indices[i*3+2] = index[2];
				// Add Normal for polygon
				Vector3f p1 = Vertices[index[0]].position;
				Vector3f p2 = Vertices[index[1]].position;
				Vector3f p3 = Vertices[index[2]].position;
				Vector3f u = p2-p1;
				Vector3f v = p3-p2;
				Vector3f n = CrossProduct(u, v);
				n.Normalize();
				Vertices[index[0]].normal += n;
				Vertices[index[1]].normal += n;
				Vertices[index[2]].normal += n;
			}
		}

		float scale = .5 * std::max(std::max(BoundingBox.MaxXSize(), BoundingBox.MaxYSize()), BoundingBox.MaxZSize());
		Vector3f center = BoundingBox.Center();
		for (int i = 0; i < numVertices; i++)
		{
			Vertices[i].position = (Vertices[i].position - center) / scale * size;
			Vertices[i].normal.Normalize();
		}

		Surfaces[0].first = 0;
		Surfaces[0].count = Indices.size();
		Surfaces[0].mode = GL_TRIANGLES;

		std::cout << "StaticModel::LoadXYZ() --- Finished Loading " << filename << "\n";

		return true;
	}


	bool StaticModel::SaveXYZ(const std::string &filename)
	{
		return true;
	}

	bool StaticModel::LoadXYZ2(const std::string &filename)
	{
		return true;
	}

	bool StaticModel::SaveXYZ2(const std::string &filename)
	{
		return true;
	}

	bool StaticModel::SaveJS(const std::string &filename, const std::string &modelName)
	{
		ofstream fout(filename);
		if (!fout)
			return false;

		int first = 0;
		int last = Surfaces.size() - 1;
		int totalVertexCount = 0;

		// Draw each surface
		fout << "mesh_" << modelName << "_vertices = [";
		for (int i = first; i <= last; ++i)
		{
			//glBegin(GL_TRIANGLES);
			fout << "[";
			for (int j = 0; j < Surfaces[i].count; ++j)
			{
				Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first + j]];
				fout << vtx->position.v[0] << "," << vtx->position.y << "," << vtx->position.z;
				if (j < Surfaces[i].count - 1) fout << ",";
				//glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
				//glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
				//glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
				//glNormal3fv(vtx->normal.v);
				//glVertex3fv(vtx->position.v);
				totalVertexCount++;
			}
			fout << "]";
			if (i != last)
				fout << ",\n";
			else
				fout << "];\n";
			//glEnd();
		}

		fout << "mesh_" << modelName << "_normals = [";
		for (int i = first; i <= last; ++i)
		{
			fout << "[";
			for (int j = 0; j < Surfaces[i].count; ++j)
			{
				Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first + j]];
				fout << vtx->normal.v[0] << "," << vtx->normal.y << "," << vtx->normal.z;
				if (j < Surfaces[i].count - 1)
					fout << ",";
			}
			fout << "]";
			if (i != last)
				fout << ",\n";
			else
				fout << "];\n";
		}

		fout << "mesh_" << modelName << "_texcoords = [";
		for (int i = first; i <= last; ++i)
		{
			fout << "[";
			for (int j = 0; j < Surfaces[i].count; ++j)
			{
				Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first + j]];
				fout << vtx->texcoord1.v[0] << "," << vtx->texcoord1.y;
				if (j < Surfaces[i].count - 1)
					fout << ",";
			}
			fout << "]";
			if (i != last)
				fout << ",\n";
			else
				fout << "];\n";
		}

		fout << "mesh_" << modelName << "_tangent = [";
		for (int i = first; i <= last; ++i)
		{
			fout << "[";
			for (int j = 0; j < Surfaces[i].count; ++j)
			{
				Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first + j]];
				fout << vtx->texcoord2.v[0] << "," << vtx->texcoord2.y << "," << vtx->texcoord2.z << "," << vtx->texcoord2.w;
				if (j < Surfaces[i].count - 1)
					fout << ",";
			}
			fout << "]";
			if (i != last)
				fout << ",\n";
			else
				fout << "];\n";
		}

		fout << "mesh_" << modelName << "_binormal = [";
		for (int i = first; i <= last; ++i)
		{
			fout << "[";
			for (int j = 0; j < Surfaces[i].count; ++j)
			{
				Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first + j]];
				fout << vtx->texcoord3.v[0] << "," << vtx->texcoord3.y << "," << vtx->texcoord3.z << "," << vtx->texcoord3.w;
				if (j < Surfaces[i].count - 1)
					fout << ",";
			}
			fout << "]";
			if (i != last)
				fout << ",\n";
			else
				fout << "];\n";
		}

		fout << "mesh_" << modelName << "_drawMode = gl.TRIANGLES;\n";
		fout << "mesh_" << modelName << "_numSurfaces = " << last << ";\n";
		fout << "mesh_" << modelName << "_numVertices = " << totalVertexCount << ";\n";

		return true;
	}

	bool StaticModel::LoadOBJ(const std::string &filename, const double size)
	{
		using namespace std;

		cout << "StaticModel::LoadOBJ() --- Loading " << filename << "\n";

		int curSurface = 0;
		string surfaceName;
		string objectName;
		string materialLibrary;
		float v[3];
		int iv[3];
		vector<Vector3f> vList;
		vector<Vector3f> vnList;
		vector<Vector2f> vtList;
		vector<TVector3<int>> faceList;
		map<long long, int> indexMap;	// this maps the unique combo of of vertices
		map<int, Vertex64> vertexMap;
		int newIndex = 0;
		int first = 0;
		int count = 0;
		int firstVertex = 0;

		BoundingBox.Reset();

		ifstream fin(filename.c_str());
		string str;
		istringstream istr;

		if (!fin)
			return false;

		CreateSimpleModel(0, 0, 1);
		int linecount = 0;
		int totallinecount = 0;
		while (1)
		{
			getline(fin, str);
			linecount++;
			totallinecount++;
			if (!fin)
			{
				// flush remaining polygons to last surface (if any)
				if (Surfaces.size() != 0)
				{
					Surfaces[curSurface].count = faceList.size() * 3;

					cout << "Adding " << faceList.size() << " new faces starting at index " << first << " to " << Surfaces[curSurface].surfaceName << ".\n";
					// 2. add indices (triangles)
					for (auto it = faceList.begin(); it != faceList.end(); it++)
					{
						Indices.push_back(it->v[0]+firstVertex);
						Indices.push_back(it->v[1]+firstVertex);
						Indices.push_back(it->v[2]+firstVertex);
					}
					first = Indices.size();
					faceList.clear();
				}
				break;
			}
			istr.clear();
			istr.str(str);

			istr >> str;
			if (str == "o")
			{
				//cout << "current line: " << totallinecount << "\t";
				//cout << "lines read: " << linecount << endl;
				linecount = 0;

				// flush old vertices/normals/texcoords
				// 1. add vertices
				//vertexMap.clear();
				//indexMap.clear();
				//vList.clear();
				//vnList.clear();
				//vtList.clear();
				//first = Indices.size();
				//firstVertex = Vertices.size();
				istr >> objectName;
				cout << "starting new object " << objectName << endl;
			}
			else if (str == "g")
			{
				//cout << "current line: " << totallinecount << "\t";
				//cout << "lines read: " << linecount << endl;
				linecount = 0;
				bool createSurface = false;

				// flush old polygons to last surface (if any)
				if (Surfaces.size() != 0 && !faceList.empty())
				{
					Surfaces[curSurface].count = faceList.size() * 3;

					cout << "Adding " << faceList.size() << " new faces starting at index " << first << " to " << Surfaces[curSurface].surfaceName << ".\n";
					// 2. add indices (triangles)
					for (auto it = faceList.begin(); it != faceList.end(); it++)
					{
						Indices.push_back(it->v[0]+firstVertex);
						Indices.push_back(it->v[1]+firstVertex);
						Indices.push_back(it->v[2]+firstVertex);
					}
					first = Indices.size();
					faceList.clear();
					createSurface = true;
				}

				if (createSurface)
				{
					// create new surface
					istr >> surfaceName;
					cout << "starting new surface " << surfaceName << endl;
					Surface surface;
					surface.first = first;
					surface.mode = GL_TRIANGLES;
					Surfaces.push_back(surface);
					curSurface = Surfaces.size()-1;
				}
				Surfaces[curSurface].surfaceName = surfaceName;
			}
			else if (str == "usemtl")
			{
				istr >> str;
				Surfaces[curSurface].materialName = str;
				cout << "using material " << str << endl;
			}
			else if (str == "mtllib")
			{
				istr >> materialLibrary;
				cout << "material library " << materialLibrary << endl;
			}
			else if (str == "v")
			{
				istr >> v[0] >> v[1] >> v[2];
				vList.push_back(Vector3f(v));
			}
			else if (str == "vn")
			{
				istr >> v[0] >> v[1] >> v[2];
				vnList.push_back(Vector3f(v));
			}
			else if (str == "vt")
			{
				istr >> v[0] >> v[1];
				vtList.push_back(Vector2f(v));
			}
			else if (str == "f")
			{
				long long i1 = 0, i2 = 0, i3 = 0;
				long long longIndex;
				char trashChar;
				for (int k = 0; k < 3; k++)
				{
					istr >> i1;
					istr >> trashChar;
					if (trashChar != '/')
					{
						// single position (v only)
						istr.putback(trashChar);
						i2 = i3 = -1;
					}
					else
					{
						istr >> trashChar;
						if (trashChar == '/')
						{
							// v//vn case
							i2 = -1;
							istr >> i3;
						}
						else
						{
							istr.putback(trashChar);
							istr >> i2 >> trashChar;
							if (trashChar != '/')
							{
								istr.putback(trashChar);
								i3 = i2;
								i2 = -1;
							}
							else
							{
								istr >> i3;
							}
						}
					}
					if (i1 < 0)
						i1 = vList.size() + i1;
					if (i2 < 0)
						i2 = vtList.size() + i2;
					if (i3 < 0)
						i3 = vnList.size() + i3;
					longIndex = i1;
					if (i2 >= 0)
						longIndex |= i2 << 48;
					if (i3 >= 0)
						longIndex |= i3 << 24;
					if (indexMap.find(longIndex) == indexMap.end())
					{
						indexMap[longIndex] = newIndex;
						newIndex++;
					}
					iv[k] = indexMap[longIndex];

					// create vertex if needed
					Vertex64 vtx;
					if (vertexMap.find(iv[k]) == vertexMap.end())
					{
						vtx.position = vList[i1-1];
						if (i3 >= 0) vtx.normal = vnList[i3-1];
						if (i2 >= 0) vtx.texcoord1 = vtList[i2-1];
						vertexMap[iv[k]] = vtx;
						BoundingBox += vtx.position;
					}
				}
				// create face
				faceList.push_back(TVector3<int>(iv));

			}
		}
		fin.close();

		float scale = 1;
		if (size < 0)
			scale = -2 * size / BoundingBox.MaxUniformScale();

		cout << "scale is " << scale << endl;
		for (auto it = vertexMap.begin(); it != vertexMap.end(); it++)
		{
			if (size < 0)
				it->second.position -= BoundingBox.Center();

			it->second.position *= scale;
			Vertices.push_back(it->second);
		}

		ComputeTangentVectors();

		cout << "StaticModel::LoadOBJ --- Finished Loading " << filename << "\n";
		return true;
	}

	bool StaticModel::SaveOBJ(const std::string &filename)
	{
		using namespace std;

		ofstream fout(filename.c_str());

		for (int i = 0; i < Vertices.size(); i++)
		{
			cout << "v " << Vertices[i].position << endl;
		}

		for (int i = 0; i < Vertices.size(); i++)
		{
			cout << "vn " << Vertices[i].normal << endl;
		}

		for (int i = 0; i < Vertices.size(); i++)
		{
			cout << "vt " << Vertices[i].texcoord1 << endl;
		}

		for (int surface = 0; surface < Surfaces.size(); surface++)
		{
			int first = Surfaces[surface].first;
			int count = Surfaces[surface].count;
			int faceCount = (count - first) / 3;
			cout << "# " << faceCount << " faces\n";
			cout << "o " << Surfaces[surface].surfaceName << endl;

			for (int f = first; f < first+count; f+=3)
			{
				cout << "f " << Indices[first+f] << Indices[first+f+1] << Indices[first+f+2] << endl;
			}
		}

		fout.close();
		return true;
	}

	void StaticModel::Transform(const ftl::Matrix4d &mat)
	{
		std::vector<Vertex64>::iterator it;
		for (it = Vertices.begin(); it != Vertices.end(); it++)
		{
			ftl::Vector3d vtx = it->position;
			it->position = (mat * ftl::Vector4d(vtx, 1)).xyz();
		}
	}

	void StaticModel::ComputeTangentVectors()
	{
		for (int i = 0; i < Vertices.size(); i++)
		{
			Vertices[i].texcoord2 = Vector4f(0, 0, 0, 0);
			Vertices[i].texcoord3 = Vector4f(0, 0, 0, 0);
		}

		for (int i = 0; i < Indices.size(); i+=3)
		{
			int indices[3];
			indices[0] = Indices[i];
			indices[1] = Indices[i+1];
			indices[2] = Indices[i+2];
			Vector3f v1, v2;
			v1 = Vertices[indices[1]].position - Vertices[indices[0]].position;
			v2 = Vertices[indices[2]].position - Vertices[indices[0]].position;
			Vector3f faceNormal = CrossProduct(v1, v2);
			faceNormal.Normalize();

			float s1, t1, s2, t2;
			s1 = Vertices[indices[1]].texcoord1.s - Vertices[indices[0]].texcoord1.s;
			t1 = Vertices[indices[1]].texcoord1.t - Vertices[indices[0]].texcoord1.t;
			s2 = Vertices[indices[2]].texcoord1.s - Vertices[indices[0]].texcoord1.s;
			t2 = Vertices[indices[2]].texcoord1.t - Vertices[indices[0]].texcoord1.t;
			Vector3f Tx, Ty, Tz;
			float coef = 1.0 / (s1 * t2 - s2 * t1);
			Tx.x = coef * (v1.x * t2 - v2.x * t1);
			Tx.y = coef * (v1.y * t2 - v2.y * t1);
			Tx.z = coef * (v1.z * t2 - v2.z * t1);
			Ty.x = coef * (v2.x * s1 - v1.x * s2);
			Ty.y = coef * (v2.y * s1 - v1.y * s2);
			Ty.z = coef * (v2.z * s1 - v1.z * s2);
			//Tz = Vertices[indices[j]].normal;
			//float t = DotProduct(faceNormal, Tz);
			//Tx = faceNormal - Tz * t;
			//Ty = CrossProduct(Tx, Tz);
			for (int j = 0; j < 3; j++)
			{
				Vertices[indices[j]].texcoord2 += Vector4f(Tx, 0);
				Vertices[indices[j]].texcoord3 += Vector4f(Ty, 0);
			}
		}

		for (int i = 0; i < Vertices.size(); i++)
		{
			Vector3f n = Vertices[i].normal;
			Vector3f t(Vertices[i].texcoord2.x, Vertices[i].texcoord2.y, Vertices[i].texcoord2.z);
			Vector3f b(Vertices[i].texcoord3.x, Vertices[i].texcoord3.y, Vertices[i].texcoord3.z);
			t.Normalize();
			b.Normalize();

			float T2 = t.LengthSquared();
			Vector3f tprime = (t - n * DotProduct(n, t));
			Vector3f bprime = (b - n * DotProduct(n, n) - tprime * DotProduct(tprime, b) / T2);

			Matrix3f M(tprime.x, tprime.y, tprime.z, bprime.x, bprime.y, bprime.z, n.x, n.y, n.z);

			Vertices[i].texcoord2 = tprime;
			Vertices[i].texcoord2.w = M.Determinant();
			Vertices[i].texcoord2.w = (DotProduct(CrossProduct(n, t), b) < 0.0f) ? -1.0f : 1.0f;
			Vertices[i].texcoord3 = bprime;
		}
	}


	void StaticModel::BindVertexArrays(ShaderProgram *shaderProgram)
	{
		if (shaderProgram == NULL)
		{
			vertexBinding = 0;
			normalBinding = 0;
			texcoord1Binding = 0;
			texcoord2Binding = 0;
			texcoord3Binding = 0;
		}
		else
		{			
			vertexBinding = shaderProgram->GetAttribIndex(shaderProgram->TranslateAttribUniformName("gl_Vertex"));
			normalBinding = shaderProgram->GetAttribIndex(shaderProgram->TranslateAttribUniformName("gl_Normal"));
			texcoord1Binding = shaderProgram->GetAttribIndex(shaderProgram->TranslateAttribUniformName("gl_MultiTexCoord0"));
			texcoord2Binding = shaderProgram->GetAttribIndex(shaderProgram->TranslateAttribUniformName("gl_MultiTexCoord1"));
			texcoord3Binding = shaderProgram->GetAttribIndex(shaderProgram->TranslateAttribUniformName("gl_MultiTexCoord2"));
		}
		if (vertexBinding >= 0)
		{
			glEnableVertexAttribArray(vertexBinding);
			glVertexAttribPointer(vertexBinding, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex64), (const GLvoid*)0);
		}
		else
		{
			glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(3, GL_FLOAT, sizeof(Vertex64), (const GLvoid*)0);
		}
		if (normalBinding >= 0)
		{
			glEnableVertexAttribArray(normalBinding);
			glVertexAttribPointer(normalBinding, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex64), (const GLvoid*)12);
		}
		else
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, sizeof(Vertex64), (const GLvoid*)12);
		}
		if (texcoord1Binding >= 0)
		{
			glEnableVertexAttribArray(texcoord1Binding);
			glVertexAttribPointer(texcoord1Binding, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex64), (const GLvoid*)24);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE0);
			glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex64), (const GLvoid*)24);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		if (texcoord2Binding >= 0)
		{
			glEnableVertexAttribArray(texcoord2Binding);
			glVertexAttribPointer(texcoord2Binding, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex64), (const GLvoid*)32);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE1);
			glTexCoordPointer(4, GL_FLOAT, sizeof(Vertex64), (const GLvoid*)32);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		if (texcoord3Binding >= 0)
		{
			glEnableVertexAttribArray(texcoord3Binding);
			glVertexAttribPointer(texcoord3Binding, 4, GL_FLOAT, GL_FALSE, sizeof(Vertex64), (const GLvoid*)48);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE2);
			glTexCoordPointer(4, GL_FLOAT, sizeof(Vertex64), (const GLvoid*)48);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	}


	void StaticModel::UnbindVertexArrays(ShaderProgram *shaderProgram)
	{
		if (vertexBinding > 0)
		{
			glDisableVertexAttribArray(vertexBinding);
		}
		else
		{
			glDisableClientState(GL_VERTEX_ARRAY);
		}

		if (normalBinding > 0)
		{
			glDisableVertexAttribArray(normalBinding);
		}
		else
		{
			glDisableClientState(GL_NORMAL_ARRAY);
		}

		if (texcoord1Binding > 0)
		{
			glDisableVertexAttribArray(texcoord1Binding);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE0);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}

		if (texcoord2Binding > 0)
		{
			glDisableVertexAttribArray(texcoord2Binding);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE1);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}

		if (texcoord3Binding > 0)
		{
			glDisableVertexAttribArray(texcoord3Binding);
		}
		else
		{
			glClientActiveTexture(GL_TEXTURE2);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		glClientActiveTexture(GL_TEXTURE0);
	}


	void StaticModel::RenderSurface(int i)
	{
		glDrawElements(Surfaces[i].mode, Surfaces[i].count, GL_UNSIGNED_INT, (const GLvoid*)(4*Surfaces[i].first));
	}


	void StaticModel::SetShaderProgram(ShaderProgram *shaderProgram, int surface)
	{
		int first = 0;
		int last = Surfaces.size() - 1;
		if (surface >= 0)
		{
			first = last = surface;
		}
		for (int s = first; s <= last; s++)
		{
			Surfaces[s].shaderProgram = shaderProgram;
		}
	}


	void StaticModel::Render(int surface, bool fixedFunction)
	{
		// Set range for individual surface drawing if applicable
		int first, last;
		if (surface < 0)
		{
			first = 0;
			last = Surfaces.size() - 1;
		}
		else
		{
			first = last = surface;
		}

		if (arrayBufferID != 0 && elementArrayBufferID != 0)
		{
			glBindBuffer(GL_ARRAY_BUFFER, arrayBufferID);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementArrayBufferID);
			ShaderProgram *current = NULL;
			if (fixedFunction == false)
			{
				ShaderProgram *current = Surfaces[0].shaderProgram;
			}			
			if (current != 0)
				current->Use();
			BindVertexArrays(current);
			for (int s = first; s <= last; s++)
			{
				if (Surfaces[s].shaderProgram != current)
				{
					current = Surfaces[s].shaderProgram;
					if (current != NULL)
						current->Use();
					else
						glUseProgram(0);
					if (s != first)
						UnbindVertexArrays(current);
					BindVertexArrays(current);
				}
				RenderSurface(s);
			}
			UnbindVertexArrays(current);
			glUseProgram(0);
			glBindBuffer(GL_ARRAY_BUFFER, 0);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		}
		else
		{
			// Draw each surface
			for (int i = first; i <= last; ++i)
			{
				glBegin(GL_TRIANGLES);
				for (int j = 0; j < Surfaces[i].count; ++j)
				{
					Vertex64 *vtx = &Vertices[Indices[Surfaces[i].first+j]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
				}
				glEnd();
			}
		}
	}


	void StaticModel::RenderLines(int surface)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		Render(surface);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}


	void StaticModel::RenderPoints(int surface)
	{
		// Set range for individual surface drawing if applicable
		int first, last;
		if (surface < 0)
		{
			glBegin(GL_POINTS);
			int numPoints = Vertices.size();
			for (int i = 0; i < numPoints; i++)
			{
				glNormal3fv(Vertices[i].normal.v);
				glVertex3fv(Vertices[i].position.v);
			}
			glEnd();
		}
		else
		{
			glBegin(GL_POINTS);
			int i1 = Surfaces[surface].first;
			int i2 = i1 + Surfaces[surface].count;
			for (int i = i1; i < i2; i++)
			{
				glNormal3fv(Vertices[Indices[i]].normal.v);
				glVertex3fv(Vertices[Indices[i]].position.v);
			}
			glEnd();
		}
	}


	void StaticModel::CreateSimpleModel(int vertexCount, int indexCount, int surfaceCount)
	{
		Clear();
		Resize(vertexCount, indexCount, surfaceCount);
	}


	void StaticModel::Resize(int vertexCount, int indexCount, int surfaceCount)
	{
		if (vertexCount > 0)
			Vertices.resize(vertexCount);
		if (indexCount > 0)
			Indices.resize(indexCount);
		if (surfaceCount > 0)
			Surfaces.resize(surfaceCount);			
	}


	void StaticModel::Clear()
	{
		Vertices.clear();
		Indices.clear();
		Surfaces.clear();
	}


	void StaticModel::Compile()
	{
		if (Vertices.size() == 0 || Indices.size() == 0)
			return;
		if (arrayBufferID == 0)
		{
			glGenBuffers(1, &arrayBufferID);
		}
		if (elementArrayBufferID == 0)
		{
			glGenBuffers(1, &elementArrayBufferID);
		}
		glBindBuffer(GL_ARRAY_BUFFER, arrayBufferID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementArrayBufferID);
		glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex64)*Vertices.size(), &Vertices[0], GL_STATIC_DRAW);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLuint)*Indices.size(), &Indices[0], GL_STATIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	}


	// STATIC METHODS

	StaticModel StaticModel::CreateSuperquadric(double n, double e, double alpha, int segments, int slices, double ax, double ay, double az)
	{
		using namespace std;

		StaticModel model;

		n = ftl::abs(n);
		e = ftl::abs(e);
		alpha = ftl::abs(alpha);
		ax = ftl::abs(ax);
		ay = ftl::abs(ay);
		az = ftl::abs(az);
		segments = ftl::abs(segments);
		slices = ftl::abs(slices);

		bool toroid = (alpha > 0) ? true : false;

		double u, du;
		double v, dv;

		u = -M_PI;
		du = 2 * M_PI / segments;
		v = toroid ? -M_PI : (-M_PI / 2);
		dv = toroid ? (2.0 * M_PI / slices) : (M_PI / slices);

		model.Resize((segments + 1) * (slices + 1), segments * slices * 6, 1);

		int lastsegment = segments;// - 1;
		int lastslice = slices;// - 1;

		int k = 0;
		if (!toroid)
		{
			for (int i = 0; i <= lastslice; i++)
			{				
				for (int j = 0; j <= lastsegment; j++)
				{				
					model.Vertices[i * (slices+1) + j].position.x = ax * sqC(v, n) * sqC(u, e);
					model.Vertices[i * (slices+1) + j].position.z = ay * sqC(v, n) * sqS(u, e);
					model.Vertices[i * (slices+1) + j].position.y = az * sqS(v, n);
					model.Vertices[i * (slices+1) + j].normal.x = sqC(v, 2 - n) * sqC(u, 2 - e);
					model.Vertices[i * (slices+1) + j].normal.z = sqC(v, 2 - n) * sqS(u, 2 - e);
					model.Vertices[i * (slices+1) + j].normal.y = sqS(v, 2 - n);
			
					v += dv;
				}
				v = -M_PI / 2.0;
				u += du;
			}
			for (int i = 0; i <= lastslice; i++)
			{				
				for (int j = 0; j <= lastsegment; j++)
				{
					if (i < segments && j < slices)
					{					
						model.Indices[k++] = i * (slices+1) + j;
						model.Indices[k++] = (i + 1) * (slices+1) + j;
						model.Indices[k++] = i * (slices+1) + j + 1;
						model.Indices[k++] = (i + 1) * (slices+1) + j;
						model.Indices[k++] = (i + 1) * (slices+1) + (j + 1);
						model.Indices[k++] = i * (slices+1) + j + 1;
					}
				}
			}
			// old algorithm
			//for (int i = 0; i <= lastslice; i++)
			//{				
			//	for (int j = 0; j <= lastsegment; j++)
			//	{				
			//		model.Vertices[i * (slices+1) + j].position.x = ax * sqC(v, n) * sqC(u, e);
			//		model.Vertices[i * (slices+1) + j].position.z = ay * sqC(v, n) * sqS(u, e);
			//		model.Vertices[i * (slices+1) + j].position.y = az * sqS(v, n);
			//		model.Vertices[i * (slices+1) + j].normal.x = sqC(v, 2 - n) * sqC(u, 2 - e);
			//		model.Vertices[i * (slices+1) + j].normal.z = sqC(v, 2 - n) * sqS(u, 2 - e);
			//		model.Vertices[i * (slices+1) + j].normal.y = sqS(v, 2 - n);
			//
			//		v += dv;
			//	}
			//	v = -M_PI / 2;
			//	u += du;
			//}
			//for (int i = 0; i <= lastslice; i++)
			//{				
			//	for (int j = 0; j <= lastsegment; j++)
			//	{
			//		if (i < segments && j < slices)
			//		{					
			//			model.Indices[k++] = i * (slices+1) + j;
			//			model.Indices[k++] = (i + 1) * (slices+1) + j;
			//			model.Indices[k++] = i * (slices+1) + j + 1;
			//			model.Indices[k++] = (i + 1) * (slices+1) + j;
			//			model.Indices[k++] = (i + 1) * (slices+1) + (j + 1);
			//			model.Indices[k++] = i * (slices+1) + j + 1;
			//		}
			//	}
			//}
		}
		else
		{
			for (int i = 0; i <= segments; i++)
			{
				for (int j = 0; j <= slices; j++)
				{				
					model.Vertices[i * (slices+1) + j].position.x = ax * sqCT(v, n, alpha) * sqC(u, e);
					model.Vertices[i * (slices+1) + j].position.z = ay * sqCT(v, n, alpha) * sqS(u, e);
					model.Vertices[i * (slices+1) + j].position.y = az * sqS(v, n);
					model.Vertices[i * (slices+1) + j].normal.x = sqC(v, 2 - n) * sqC(u, 2 - e);
					model.Vertices[i * (slices+1) + j].normal.z = sqC(v, 2 - n) * sqS(u, 2 - e);
					model.Vertices[i * (slices+1) + j].normal.y = sqS(v, 2 - n);
					if (i < segments && j < slices)
					{					
						model.Indices[k++] = i * (slices+1) + j;
						model.Indices[k++] = (i + 1) * (slices+1) + j;
						model.Indices[k++] = i * (slices+1) + j + 1;
						model.Indices[k++] = (i + 1) * (slices+1) + j;
						model.Indices[k++] = (i + 1) * (slices+1) + (j + 1);
						model.Indices[k++] = i * (slices+1) + j + 1;
					}
			
					v += dv;
				}
				v = -M_PI;
				u += du;
			}
		}
		//if (n == e == 0.0)
		//{
		//	for (int i = 0; i < model.Vertices.size(); i++)
		//	{
		//		cout << i << " " << model.Vertices[i].position << endl;
		//	}
		//}

		return model;
	}


	StaticModel StaticModel::CreateBox(float x1, float y1, float z1, float x2, float y2, float z2)
	{
		using namespace Fluxions;
		using namespace std;

		float minX = ftl::min(x1, x2);
		float maxX = ftl::max(x1, x2);
		float minY = ftl::min(y1, y2);
		float maxY = ftl::max(y1, y2);
		float minZ = ftl::min(z1, z2);
		float maxZ = ftl::max(z1, z2);
	
		StaticModel staticModel;
		staticModel.CreateSimpleModel(24, 36, 1);
		int i = 0;
		staticModel.Vertices[i+0].normal = Vector3f(0, -1, 0);
		staticModel.Vertices[i+1].normal = Vector3f(0, -1, 0);
		staticModel.Vertices[i+2].normal = Vector3f(0, -1, 0);
		staticModel.Vertices[i+3].normal = Vector3f(0, -1, 0);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+0].position = Vector3f(minX, minY, minZ);
		staticModel.Vertices[i+1].position = Vector3f(maxX, minY, minZ);
		staticModel.Vertices[i+2].position = Vector3f(maxX, minY, maxZ);
		staticModel.Vertices[i+3].position = Vector3f(minX, minY, maxZ);

		i += 4;
	
		staticModel.Vertices[i+3].normal = Vector3f(0, 1, 0);
		staticModel.Vertices[i+2].normal = Vector3f(0, 1, 0);
		staticModel.Vertices[i+1].normal = Vector3f(0, 1, 0);
		staticModel.Vertices[i+0].normal = Vector3f(0, 1, 0);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+3].position = Vector3f(minX, maxY, minZ);
		staticModel.Vertices[i+2].position = Vector3f(maxX, maxY, minZ);
		staticModel.Vertices[i+1].position = Vector3f(maxX, maxY, maxZ);
		staticModel.Vertices[i+0].position = Vector3f(minX, maxY, maxZ);

		i += 4;
	
		staticModel.Vertices[i+3].normal = Vector3f(-1, 0, 0);
		staticModel.Vertices[i+2].normal = Vector3f(-1, 0, 0);
		staticModel.Vertices[i+1].normal = Vector3f(-1, 0, 0);
		staticModel.Vertices[i+0].normal = Vector3f(-1, 0, 0);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+3].position = Vector3f(minX, minY, minZ);
		staticModel.Vertices[i+2].position = Vector3f(minX, maxY, minZ);
		staticModel.Vertices[i+1].position = Vector3f(minX, maxY, maxZ);
		staticModel.Vertices[i+0].position = Vector3f(minX, minY, maxZ);

		i += 4;
	
		staticModel.Vertices[i+0].normal = Vector3f(1, 0, 0);
		staticModel.Vertices[i+1].normal = Vector3f(1, 0, 0);
		staticModel.Vertices[i+2].normal = Vector3f(1, 0, 0);
		staticModel.Vertices[i+3].normal = Vector3f(1, 0, 0);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+0].position = Vector3f(maxX, minY, minZ);
		staticModel.Vertices[i+1].position = Vector3f(maxX, maxY, minZ);
		staticModel.Vertices[i+2].position = Vector3f(maxX, maxY, maxZ);
		staticModel.Vertices[i+3].position = Vector3f(maxX, minY, maxZ);

		i += 4;
	
		staticModel.Vertices[i+3].normal = Vector3f(0, 0, -1);
		staticModel.Vertices[i+2].normal = Vector3f(0, 0, -1);
		staticModel.Vertices[i+1].normal = Vector3f(0, 0, -1);
		staticModel.Vertices[i+0].normal = Vector3f(0, 0, -1);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+3].position = Vector3f(minX, minY, minZ);
		staticModel.Vertices[i+2].position = Vector3f(maxX, minY, minZ);
		staticModel.Vertices[i+1].position = Vector3f(maxX, maxY, minZ);
		staticModel.Vertices[i+0].position = Vector3f(minX, maxY, minZ);

		i += 4;
	
		staticModel.Vertices[i+0].normal = Vector3f(0, 0, 1);
		staticModel.Vertices[i+1].normal = Vector3f(0, 0, 1);
		staticModel.Vertices[i+2].normal = Vector3f(0, 0, 1);
		staticModel.Vertices[i+3].normal = Vector3f(0, 0, 1);
		staticModel.Vertices[i+0].texcoord1 = Vector2f(0, 1);
		staticModel.Vertices[i+1].texcoord1 = Vector2f(1, 1);
		staticModel.Vertices[i+2].texcoord1 = Vector2f(1, 0);
		staticModel.Vertices[i+3].texcoord1 = Vector2f(0, 0);
		staticModel.Vertices[i+0].position = Vector3f(minX, minY, maxZ);
		staticModel.Vertices[i+1].position = Vector3f(maxX, minY, maxZ);
		staticModel.Vertices[i+2].position = Vector3f(maxX, maxY, maxZ);
		staticModel.Vertices[i+3].position = Vector3f(minX, maxY, maxZ);

		int j = 0;
		for (i = 0; i < 6; i++)
		{
			j = i*6;
			staticModel.Indices[j+0] = i*4+0;
			staticModel.Indices[j+1] = i*4+1;
			staticModel.Indices[j+2] = i*4+2;
			staticModel.Indices[j+3] = i*4+2;
			staticModel.Indices[j+4] = i*4+3;
			staticModel.Indices[j+5] = i*4+0;
		}

		staticModel.Surfaces[0].mode = GL_QUADS;
		staticModel.Surfaces[0].first = 0;
		staticModel.Surfaces[0].count = 36;
		staticModel.Surfaces[0].materialName = "bump_perpixel";

		staticModel.ComputeTangentVectors();

		return staticModel;
	}


	void StaticModel::DebugRender(int surface, GLenum drawMode,
		bool drawNormals, bool useTexture, bool useLight)
	{
		// Set range for individual surface drawing if applicable
		int first, last;
		if (surface < 0)
		{
			first = 0;
			last = Surfaces.size() - 1;
		}
		else
		{
			first = last = surface;
		}
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		if (drawNormals)
		{
			for (int i = first; i <= last; i++)
			{
				glBegin(GL_LINES);
				for (int j = Surfaces[i].first; j < Surfaces[i].count; j+=3)
				{
					Vertex64 *vtx = &Vertices[Indices[j]];
					glColor3f(0, 0, 1);
					glVertex3fv(vtx->position.v);
					glVertex3fv((vtx->position + vtx->normal * 0.03f).v);
					glColor3f(1, 0, 0);
					glVertex3fv(vtx->position.v);
					glVertex3fv((vtx->position + vtx->texcoord2.xyz() * 0.03f).v);
					glColor3f(0, 1, 0);
					glVertex3fv(vtx->position.v);
					glVertex3fv((vtx->position + vtx->texcoord3.xyz() * 0.03f).v);
				}
				glEnd();
			}
		}

		if (useTexture)
		{
			glEnable(GL_TEXTURE_2D);
		}

		if (useLight)
		{
			glEnable(GL_LIGHTING);
			glEnable(GL_LIGHT0);
		}

		if (drawMode == GL_FILL)
		{
			// Draw each surface
			for (int i = first; i <= last; ++i)
			{
				glDrawElements(Surfaces[i].mode, Surfaces[i].count, GL_UNSIGNED_INT, &Indices[Surfaces[i].first]);
				//glBegin(GL_TRIANGLES);
				//for (int j = Surfaces[i].first; j < Surfaces[i].count; ++j)
				//{
				//	Vertex64 *vtx = &Vertices[Indices[j]];
				//	glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
				//	glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
				//	glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
				//	glNormal3fv(vtx->normal.v);
				//	glVertex3fv(vtx->position.v);
				//}
				//glEnd();
			}
		}
		else if (drawMode == GL_LINE)
		{
			// Draw each surface
			for (int i = first; i <= last; ++i)
			{
				//glDrawElements(Surfaces[i].mode, Surfaces[i].count, GL_UNSIGNED_SHORT, &indices[Surfaces[i].first]);
				glBegin(GL_LINES);
				for (int j = Surfaces[i].first; j < Surfaces[i].count; j+=3)
				{
					Vertex64 *vtx;
					vtx = &Vertices[Indices[j]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
					vtx = &Vertices[Indices[j+1]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
					vtx = &Vertices[Indices[j+1]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
					vtx = &Vertices[Indices[j+2]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
					vtx = &Vertices[Indices[j+2]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
					vtx = &Vertices[Indices[j]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
				}
				glEnd();
			}
		}
		else if (drawMode == GL_POINT)
		{
			// Draw each surface
			for (int i = first; i <= last; ++i)
			{
				//glDrawElements(Surfaces[i].mode, Surfaces[i].count, GL_UNSIGNED_SHORT, &indices[Surfaces[i].first]);
				glBegin(GL_POINTS);
				for (int j = Surfaces[i].first; j < Surfaces[i].count; ++j)
				{
					Vertex64 *vtx = &Vertices[Indices[j]];
					glMultiTexCoord2fv(GL_TEXTURE0, vtx->texcoord1.v);
					glMultiTexCoord4fv(GL_TEXTURE1, vtx->texcoord2.v);
					glMultiTexCoord4fv(GL_TEXTURE2, vtx->texcoord3.v);
					glNormal3fv(vtx->normal.v);
					glVertex3fv(vtx->position.v);
				}
				glEnd();
			}
		}

		if (useTexture)
		{
			glDisable(GL_TEXTURE_2D);
		}

		if (useLight)
		{
			glDisable(GL_LIGHT0);
			glDisable(GL_LIGHTING);
		}		
	}
}
