#ifndef XOR_WIDGET_H
#define XOR_WIDGET_H


#include <string>
#include <memory>


using namespace std;


#define RENDER_PASS0	0
#define RENDER_PASS1	1
#define RENDER_PASS2	2
#define RENDER_PASS3	3
#define RENDER_PASS4	4
#define RENDER_PASS5	5
#define RENDER_PASS6	6
#define RENDER_PASS7	7
#define RENDER_OVERLAY	1000


enum WidgetState
{
	WS_NotRunning,
	WS_Foreground,
	WS_Background,
	WS_Suspended
};


enum WidgetRunningState
{
	WRS_Running,
	WRS_Paused,
	WRS_Stopped
};


// A Two 
class Widget
{
private:
	WidgetState lastState = WS_NotRunning;
	WidgetState currentState = WS_NotRunning;
	WidgetRunningState lastRunningState = WRS_Stopped;
	WidgetRunningState currentRunningState = WRS_Stopped;

	string name;

public:
	Widget() { name = "Widget"; }
	Widget(const string &widgetName) { name = widgetName; }
	virtual ~Widget() { }

	void setName(const string &name) { this->name = name; }
	const string &getName() const { return name; }
	const WidgetState getState() const { return currentState; }
	const WidgetRunningState getRunningState() const { return currentRunningState; }

	bool isForeground() const { return currentState == WS_Foreground; }
	bool isBackground() const { return currentState == WS_Background; }
	bool isSuspended() const { return currentState == WS_Suspended; }
	bool isRunning() const { return currentRunningState == WRS_Running; }
	bool isPaused() const { return currentRunningState == WRS_Paused; }
	bool isStopped() const { return currentRunningState == WRS_Stopped; }

	void SwitchState(WidgetState newState) { lastState = currentState; currentState = newState; }
	void SwitchRunningState(WidgetRunningState newState) { lastRunningState = currentRunningState; currentRunningState = newState; }

	virtual void Init() { OnInit(); }
	virtual void Kill() { OnKill(); }

	virtual void Start() { SwitchRunningState(WRS_Running);  OnStart(); }
	virtual void Stop() { SwitchRunningState(WRS_Stopped); OnStop(); }
	virtual void Pause() { SwitchRunningState(WRS_Paused); OnPause(); }
	virtual void Resume() { SwitchRunningState(WRS_Running); OnResume(); }

	void LeaveState()
	{
		if (isSuspended())	OnLeaveSuspended();
		if (isForeground())	OnLeaveForeground();
		if (isBackground()) OnLeaveBackground();
		SwitchState(WS_NotRunning);
	}

	void SwitchToSuspended() { LeaveState(); SwitchState(WS_Suspended); OnEnterSuspended(); }
	void SwitchToBackground() {	LeaveState(); SwitchState(WS_Background); OnEnterBackground(); }
	void SwitchToForeground() { LeaveState(); SwitchState(WS_Foreground); OnEnterForeground(); }

	// Event Handlers

	virtual void OnInit() { }
	virtual void OnKill() { }
	virtual void OnStart() { }
	virtual void OnStop() { }
	virtual void OnPause() { }
	virtual void OnResume() { }
	
	virtual void OnIdle(double deltaTime)
	{
		if (!isRunning())
			return;

		// Ignore all other states beside Foreground and Background...
		if (isForeground())
		{
			OnIdleForeground(deltaTime);
		}
		else if (isBackground())
		{
			OnIdleBackground(deltaTime);
		}
	}

	virtual void OnIdleForeground(double deltaTime) { }
	virtual void OnEnterForeground() { }
	virtual void OnLeaveForeground() { }

	virtual void OnIdleBackground(double deltaTime) { }
	virtual void OnEnterBackground() { }
	virtual void OnLeaveBackground() { }

	virtual void OnEnterSuspended() { }
	virtual void OnLeaveSuspended() { }

	virtual void OnJoystickAxis(int axis, float value) { }
	virtual void OnJoystickButtonDown(int button) { }
	virtual void OnJoystickButtonUp(int button) { }
	virtual void OnKeyDown(int key) { }
	virtual void OnKeyUp(int key) { }
	virtual void OnKeyChar(int unicode) { }
	virtual void OnMouseMotion(int x, int y, int dx, int dy) { }
	virtual void OnMouseButtonDown(int x, int y, int button) { }
	virtual void OnMouseButtonUp(int x, int y, int button) { }
	virtual void OnDisplayResize(int width, int height) { }

	virtual void OnRender(int pass) { }
};


class WidgetContainer : public Widget
{
public:
	vector<Widget*> children;

	WidgetContainer() { setName("WidgetContainer"); }
	~WidgetContainer() { }

	int size() const { return children.size(); }
	Widget *FindFirstWidget(const string &name);
	Widget *operator[](const string &name) { return FindFirstWidget(name); }
	Widget *operator[](size_t i) { if (i >= children.size()) return NULL; return children[i]; }

	virtual void OnInit();
	virtual void OnKill();
	virtual void OnStart();
	virtual void OnStop();
	virtual void OnPause();
	virtual void OnResume();

	virtual void OnIdle(double deltaTime);
	virtual void OnIdleForeground(double deltaTime);
	virtual void OnEnterForeground();
	virtual void OnLeaveForeground();
	virtual void OnIdleBackground(double deltaTime);
	virtual void OnEnterBackground();
	virtual void OnLeaveBackground();
	virtual void OnEnterSuspended();
	virtual void OnLeaveSuspended();
	virtual void OnJoystickAxis(int axis, float value);
	virtual void OnJoystickButtonDown(int button);
	virtual void OnJoystickButtonUp(int button);
	virtual void OnKeyDown(int key);
	virtual void OnKeyUp(int key);
	virtual void OnKeyChar(int unicode);
	virtual void OnMouseMotion(int x, int y, int dx, int dy);
	virtual void OnMouseButtonDown(int x, int y, int button);
	virtual void OnMouseButtonUp(int x, int y, int button);
	virtual void OnDisplayResize(int width, int height);

	virtual void OnRender(int pass);
};


class WidgetSwitch : public WidgetContainer
{
public:
	int selectedChild = -1;
	bool actLikeContainer = true;

	WidgetSwitch() { setName("WidgetSwitch"); }
	~WidgetSwitch() { }

	void setChild(int child) {
		if (child < 0 || child >= size())
			selectedChild = -1;
		else
			selectedChild = child;
	}

	Widget *getSelected() {
		if (selectedChild >= 0 && selectedChild < size())
			return (*this)[selectedChild];
		return NULL;
	}

	virtual void OnInit();
	virtual void OnKill();
	virtual void OnStart();
	virtual void OnStop();
	virtual void OnPause();
	virtual void OnResume();

	virtual void OnIdle(double deltaTime);
	virtual void OnIdleForeground(double deltaTime);
	virtual void OnEnterForeground();
	virtual void OnLeaveForeground();
	virtual void OnIdleBackground(double deltaTime);
	virtual void OnEnterBackground();
	virtual void OnLeaveBackground();
	virtual void OnEnterSuspended();
	virtual void OnLeaveSuspended();
	virtual void OnJoystickAxis(int axis, float value);
	virtual void OnJoystickButtonDown(int button);
	virtual void OnJoystickButtonUp(int button);
	virtual void OnKeyDown(int key);
	virtual void OnKeyUp(int key);
	virtual void OnKeyChar(int unicode);
	virtual void OnMouseMotion(int x, int y, int dx, int dy);
	virtual void OnMouseButtonDown(int x, int y, int button);
	virtual void OnMouseButtonUp(int x, int y, int button);
	virtual void OnDisplayResize(int width, int height);

	virtual void OnRender(int pass);
};


#endif
